#pragma once

#include <cmath>

#include "core/tensor.h"
#include "kern/kernel_define.h"
#include "kern/naive/naive.h"
#include "kern/naive/quantize.h"

#include "common.h"

namespace inferllm {
namespace opt {
#if 1
inline void dequantize_row_q4_0(const void* __restrict x, float* __restrict y, int k) {
    const int nb = k / QK40;
    const BlockQ40* vx = static_cast<const BlockQ40*>(x);

    // scalar
    for (int i = 0; i < nb; i++) {
        const float d = vx[i].d;

        const uint8_t* pp = vx[i].qs;

        for (int l = 0; l < QK40; l += 2) {
            const uint8_t vi = pp[l / 2];

            const int8_t vi0 = vi & 0xf;
            const int8_t vi1 = vi >> 4;

            const float v0 = (vi0 - 8) * d;
            const float v1 = (vi1 - 8) * d;

            y[i * QK40 + l + 0] = v0;
            y[i * QK40 + l + 1] = v1;
        }
    }
}

inline void quantize_row_q8_0(const float* __restrict x, void* __restrict vy, int k) {
    BlockQ80* y = static_cast<BlockQ80*>(vy);

    const int nb = k / QK80;

    for (int i = 0; i < nb; i++) {
        float amax = vmaxabs(QK80, &x[i * QK80], 0.0);

        const float d = amax / ((1 << 7) - 1);
        const float id = d ? 1.0f / d : 0.0f;

        y[i].d = d;

        for (int l = 0; l < QK80; ++l) {
            const float v0 = x[i * QK80 + l] * id;

            y[i].qs[l] = roundf(v0);
        }
    }
}
#endif

inline void quantize_row_q8_0_reference(const float* x, BlockQ80* y, int k) {
    assert(k % QK80 == 0);
    const int nb = k / QK80;

    for (int i = 0; i < nb; i++) {
        float amax = 0.0f;  // absolute max

        for (int l = 0; l < QK80; l++) {
            const float v = x[i * QK80 + l];
            amax = std::max(amax, fabsf(v));
        }

        const float d = amax / ((1 << 7) - 1);
        const float id = d ? 1.0f / d : 0.0f;

        y[i].d = d;

        for (int l = 0; l < QK80; ++l) {
            const float v0 = x[i * QK80 + l] * id;

            y[i].qs[l] = roundf(v0);
        }
    }
}

inline void dequantize_row_q8_0_reference(
        const void* __restrict x, float* __restrict y, int k) {
    const int nb = k / QK80;
    const size_t bs = sizeof(float) + QK40 / 2;

    const BlockQ80* xx = reinterpret_cast<const BlockQ80*>(x);

    // scalar
    for (int i = 0; i < nb; i++) {
        const float d = xx[i].d;

        const int8_t* __restrict pp = xx[i].qs;

        for (int l = 0; l < QK80; l++) {
            y[i * QK80 + l] = pp[l] * d;
        }
    }
}

inline float vec_vec_dot_q40_with_q80_reference(
        const int n, const void* vx, const void* vy) {
    const int nb = n / QK80;
    assert(n % QK80 == 0);
    assert(nb % 2 == 0);

    const BlockQ40* __restrict x = (BlockQ40*)(vx);
    const BlockQ80* __restrict y = (BlockQ80*)(vy);

    float sumf = 0.0;
    for (int i = 0; i < nb; i++) {
        const float d0 = x[i].d;
        const float d1 = y[i].d;

        const uint8_t* __restrict p0 = x[i].qs;
        const int8_t* __restrict p1 = y[i].qs;

        int sumi = 0;
        for (int j = 0; j < QK80 / 2; j++) {
            const uint8_t v0 = p0[j];

            const int i0 = (int8_t)(v0 & 0x0F) - 8;
            const int i1 = (int8_t)(v0 >> 4) - 8;

            const int i2 = p1[2 * j + 0];
            const int i3 = p1[2 * j + 1];

            sumi += i0 * i2 + i1 * i3;
        }
        sumf += d0 * d1 * sumi;
    }
    return sumf;
}
inline float vec_vec_dot_q40_with_q80(
        const int n, const void* vx, const void* vy) {
    const int nb = n / QK80;
    assert(n % QK80 == 0);
    assert(nb % 2 == 0);

    const struct BlockQ40*  x = (struct BlockQ40*)(vx);
    const struct BlockQ80*  y = (struct BlockQ80*)(vy);

    float sumf = 0.0;
    for (int i = 0; i < nb; i++) {
        const float d0 = x[i].d;
        const float d1 = y[i].d;

        const uint8_t*  p0 = x[i].qs;
        const int8_t*  p1 = y[i].qs;

        int sumi = 0;
        asm volatile(
           "li             t1, 8              \n\t"
           "vsetvli        t0, zero, e8, mf2  \n\t"
           "vle8.v         v0, (%[X])         \n\t"
           "vwaddu.vx      v2, v0, zero       \n\t"
           "vsetvli        t0, zero, e16, m1  \n\t"
           "vwaddu.vx      v4, v2, zero       \n\t"
           "vsetvli        t0, zero, e32, m2  \n\t"
           "vand.vi        v8, v4, 0xf        \n\t"   // low  4bit
           "vsrl.vi        v0, v4, 4          \n\t"   // high 4bit
           "vsll.vi        v12, v0, 16        \n\t"
           "vor.vv         v8, v8, v12        \n\t"

           "vmv.v.i        v20, 0             \n\t"
           "vsetvli        t0, zero, e8, m1   \n\t"
           "vle8.v         v0, (%[Y])         \n\t"
           "vwadd.vx       v4, v0, zero       \n\t"
           "vsetvli        t0, zero, e16, m2  \n\t"
           "vsub.vx        v12, v8, t1        \n\t"
           "vwmul.vv       v16, v12, v4       \n\t"
           "vsetvli        t0, zero, e32, m2  \n\t"
           "vadd.vv        v8, v16, v18       \n\t"
           "vsetvli        t0, zero, e32, m1  \n\t"
           "vadd.vv        v4, v8, v9         \n\t"
           "vredsum.vs     v6, v4, v20        \n\t"
           "vmv.x.s        %[SUM], v6         \n\t"
           :[X]"+r"(p0), [Y]"+r"(p1), [SUM]"+r"(sumi)
           :
           :"cc", "t0", "t1");
        sumf += d0 * d1 * sumi;
    }
    return sumf;
}

}  // namespace opt
}  // namespace inferllm
